import { Client } from '@elastic/elasticsearch';
import { Injectable } from '@nestjs/common';
import Redis from 'ioredis';
import { DataSource } from 'typeorm';
import {
  BOOLEAN,
  BlockUserEntity,
  CardTagEntity,
  ConversationUserBackgroundEntity,
  ConversationUserDisableNotifyEntity,
  ConversationUserHiddenEntity,
  ConversationUserPinnedEntity,
  FRIEND_TYPE,
  FriendEntity,
  IMedia,
  REPORT_TYPE,
  RedisKey,
  ReportEntity,
  UserEntity,
  UserReportEntity,
  UserUnfollowNewsfeedEntity,
} from '../main';

export interface IUserProfileRedis {
  id: number;
  avatar: string;
  name: string;
}

@Injectable()
export class RedisHelper {
  private REDIS_KEY: RedisKey;

  private readonly SECONDS_FROM_HOUR = 24 * 3600;

  constructor(
    private readonly dataSource: DataSource,
    nodeEnv: string,
    private readonly redis: Redis,
    private readonly elasticIndex: {
      techres_customers: string;
    },
    private readonly elastic: Client
  ) {
    this.REDIS_KEY = new RedisKey(nodeEnv);
  }

  //====================CHECK_FRIEND_TYPE====================
  async checkFriendType(currentUserId: string, targetUserId: string) {
    if (currentUserId == targetUserId) return FRIEND_TYPE.ITS_ME;

    const hasExist = await this.redis.exists([
      this.REDIS_KEY.USER_INFO(currentUserId, 'FRIEND'),
      this.REDIS_KEY.USER_INFO(currentUserId, 'WAITING_CONFIRM'),
      this.REDIS_KEY.USER_INFO(currentUserId, 'WAITING_RESPONSE'),
    ]);
    if (hasExist !== 3) {
      await this.syncFriendDbToRedis(currentUserId);
      await this.syncWaitingConfirmDbToRedis(currentUserId);
      await this.syncWaitingResponseDbToRedis(currentUserId);
    }

    if (await this.redis.sismember(this.REDIS_KEY.USER_INFO(currentUserId, 'FRIEND'), targetUserId)) return FRIEND_TYPE.FRIEND;
    if (await this.redis.sismember(this.REDIS_KEY.USER_INFO(currentUserId, 'WAITING_CONFIRM'), targetUserId))
      return FRIEND_TYPE.WAITING_CONFIRM;
    if (await this.redis.sismember(this.REDIS_KEY.USER_INFO(currentUserId, 'WAITING_RESPONSE'), targetUserId))
      return FRIEND_TYPE.WAITING_RESPONSE;

    return FRIEND_TYPE.NOT_FRIEND;
  }

  async syncWaitingConfirmDbToRedis(syncUserId: string) {
    const userIdFriend = await this.dataSource
      .getRepository(FriendEntity)
      .find({
        where: { user_id: syncUserId, type: FRIEND_TYPE.WAITING_CONFIRM },
        order: {
          user_friend_id: 'ASC',
        },
      })
      .then((friend) =>
        friend.reduce(
          (result: string[], item) => {
            result.push(item.user_friend_id);

            return result;
          },
          ['0']
        )
      );

    await this.redis.sadd(this.REDIS_KEY.USER_INFO(syncUserId, 'WAITING_CONFIRM'), userIdFriend);
    await this.redis.expire(this.REDIS_KEY.USER_INFO(syncUserId, 'WAITING_CONFIRM'), this.SECONDS_FROM_HOUR);
  }

  async syncWaitingResponseDbToRedis(syncUserId: string) {
    const userIdFriend = await this.dataSource
      .getRepository(FriendEntity)
      .find({
        where: {
          user_id: syncUserId,
          type: FRIEND_TYPE.WAITING_RESPONSE,
        },
        order: {
          user_friend_id: 'ASC',
        },
      })
      .then((friend) =>
        friend.reduce(
          (result: string[], item) => {
            result.push(item.user_friend_id);

            return result;
          },
          ['0']
        )
      );

    await this.redis.sadd(this.REDIS_KEY.USER_INFO(syncUserId, 'WAITING_RESPONSE'), userIdFriend);
    await this.redis.expire(this.REDIS_KEY.USER_INFO(syncUserId, 'WAITING_RESPONSE'), this.SECONDS_FROM_HOUR);
  }

  //====================FIND_MUTUAL_FRIEND====================
  async findMutualFriend(currentUserId: string, targetUserId: string) {
    if (currentUserId == targetUserId) return [];

    const hasExist = await this.redis.exists([
      this.REDIS_KEY.USER_INFO(currentUserId, 'FRIEND'),
      this.REDIS_KEY.USER_INFO(targetUserId, 'FRIEND'),
    ]);
    if (hasExist != 2) {
      await this.syncFriendDbToRedis(currentUserId);
      await this.syncFriendDbToRedis(targetUserId);
    }

    const listMutualFriend: string[] = await this.redis.sinter([
      this.REDIS_KEY.USER_INFO(currentUserId, 'FRIEND'),
      this.REDIS_KEY.USER_INFO(targetUserId, 'FRIEND'),
    ]);

    listMutualFriend.splice(0, 1);

    return listMutualFriend;
  }

  //===================USER_ID_BLOCK_LIST===================
  async userIdBlockList(userId: string) {
    const hasExist = await this.redis.exists(this.REDIS_KEY.USER_INFO(userId, 'BLOCK'));
    if (!hasExist) return await this.syncBlockUserDbToRedis(userId);

    const blockList = await this.redis.smembers(this.REDIS_KEY.USER_INFO(userId, 'BLOCK'));
    blockList.splice(0, 1);

    return blockList;
  }

  async syncBlockUserDbToRedis(syncUserId: string) {
    /**
     * Lấy userId của user mình block và block mình
     */
    const userIdBlock = await this.dataSource
      .getRepository(BlockUserEntity)
      .find({
        select: ['user_id', 'block_user_id'],
        where: [
          {
            user_id: syncUserId,
          },
          {
            block_user_id: syncUserId,
          },
        ],
      })
      .then((data) =>
        data.reduce(
          (result, item) => {
            result.push(...[item.user_id, item.block_user_id].filter((item) => item != syncUserId));

            return result;
          },
          ['0']
        )
      );

    this.redis.sadd(this.REDIS_KEY.USER_INFO(syncUserId, 'BLOCK'), userIdBlock);
    this.redis.expire(this.REDIS_KEY.USER_INFO(syncUserId, 'BLOCK'), this.SECONDS_FROM_HOUR);

    userIdBlock.splice(0, 1);

    return userIdBlock;
  }

  //====================USER_UNFOLLOW_NEWSFEED====================
  async userUnfollowNewsfeed(userId: string) {
    const hasExist = await this.redis.exists(this.REDIS_KEY.USER_INFO(userId, 'UN_FOLLOW'));
    if (!hasExist) return await this.syncUserUnfollowNewsfeed(userId);

    const unfollowNewsfeedList = await this.redis.smembers(this.REDIS_KEY.USER_INFO(userId, 'UN_FOLLOW'));
    unfollowNewsfeedList.splice(0, 1);

    return unfollowNewsfeedList;
  }

  async syncUserUnfollowNewsfeed(syncUserId: string) {
    const userUnfollowNewsfeed = await this.dataSource
      .getRepository(UserUnfollowNewsfeedEntity)
      .find({
        select: ['unfollow_user_id'],
        where: {
          user_id: syncUserId,
        },
      })
      .then((data) =>
        data.reduce(
          (result, item) => {
            result.push(item.unfollow_user_id);

            return result;
          },
          ['0']
        )
      );

    this.redis.sadd(this.REDIS_KEY.USER_INFO(syncUserId, 'UN_FOLLOW'), userUnfollowNewsfeed);
    this.redis.expire(this.REDIS_KEY.USER_INFO(syncUserId, 'UN_FOLLOW'), this.SECONDS_FROM_HOUR);

    userUnfollowNewsfeed.splice(0, 1);

    return userUnfollowNewsfeed;
  }

  //====================CARD_TAG_LIST====================
  async cardTagList(cardTagIds: number[]) {
    const redisKeyCardTagList = cardTagIds.map((cardTagId) => this.REDIS_KEY.CARD_TAG(`${cardTagId}`));
    const cardTagList = await this.redis.mget(...redisKeyCardTagList);

    const result: { [card_tag_id: string]: CardTagEntity } = {};
    await Promise.all(
      cardTagList.map(async (item, index) => {
        if (!item) {
          const cardTagSync = await this.syncCardTagDbToRedis(cardTagIds[index]);

          result[cardTagIds[index]] = cardTagSync;
        } else {
          const cardTag = JSON.parse(item);
          result[cardTag.id] = cardTag;
        }
      })
    );

    return result;
  }

  async syncCardTagDbToRedis(cardTagId: number) {
    const cardTag = await this.dataSource.getRepository(CardTagEntity).findOne({
      where: {
        id: String(cardTagId),
        is_deleted: BOOLEAN.FALSE,
      },
    });
    if (!cardTag) return;

    this.redis.set(this.REDIS_KEY.CARD_TAG(String(cardTagId)), JSON.stringify(cardTag), 'EX', this.SECONDS_FROM_HOUR);

    return cardTag;
  }

  //====================POST_ID_HIDDEN_LIST====================
  async postIdHiddenList(userId: string) {
    const hasExist = await this.redis.exists(this.REDIS_KEY.USER_INFO(userId, 'HIDDEN_POST'));
    if (!hasExist) return await this.syncPostIdHiddenDbToRedis(userId);

    const postHidden = await this.redis.smembers(this.REDIS_KEY.USER_INFO(userId, 'HIDDEN_POST'));
    postHidden.splice(0, 1);
    return postHidden;
  }

  async syncPostIdHiddenDbToRedis(syncUserId: string) {
    const postIdHiddenList = await this.dataSource
      .getRepository(UserReportEntity)
      .createQueryBuilder('ur')
      .select('r.content_id')
      .leftJoinAndSelect(ReportEntity, 'r', 'ur.report_id = r.id')
      .where('ur.user_id = :userId', {
        userId: syncUserId,
      })
      .andWhere('r.type = :type', {
        type: REPORT_TYPE.POST,
      })
      .getRawMany<{ content_id: string }>()
      .then((data) =>
        data.reduce(
          (result, item) => {
            result.push(item.content_id);

            return result;
          },
          ['0']
        )
      );

    this.redis.sadd(this.REDIS_KEY.USER_INFO(syncUserId, 'HIDDEN_POST'), postIdHiddenList);
    this.redis.expire(this.REDIS_KEY.USER_INFO(syncUserId, 'HIDDEN_POST'), this.SECONDS_FROM_HOUR);

    postIdHiddenList.splice(0, 1);

    return postIdHiddenList;
  }

  //====================USER_PROFILE_LIST====================
  async userProfileList(userIdList: string[]) {
    const redisKeyUserList = userIdList.map((userId) => this.REDIS_KEY.USER_INFO(userId, 'PROFILE'));
    const userList = await this.redis.mget(...redisKeyUserList);

    const userIdSyncRedis: string[] = [];

    const usersRedis = userList.reduce((accumulator, currentValue, currentIndex) => {
      const user = JSON.parse(currentValue) as IUserProfileRedis;

      const userId = userIdList[currentIndex];
      if (!user) {
        userIdSyncRedis.push(userId);
        accumulator[userId] = {} as IUserProfileRedis;
      } else {
        accumulator[userId] = user;
      }

      return accumulator;
    }, {} as Record<string, IUserProfileRedis>);

    const usersSyncRedis = userIdSyncRedis.length ? await this.syncUserDbToRedis(userIdSyncRedis) : {};

    return { ...usersRedis, ...usersSyncRedis };
  }

  async syncUserDbToRedis(userIds: string[]) {
    const redisMulti = this.redis.multi();

    const users = await this.elastic
      .search<UserEntity>({
        index: this.elasticIndex.techres_customers,
        _source: ['id', 'avatar', 'name'],
        query: {
          bool: {
            must: [
              {
                terms: {
                  id: userIds,
                },
              },
            ],
          },
        },
      })
      .then((result) =>
        result.hits.hits.reduce((accumulator, currentValue) => {
          const user = currentValue._source;

          accumulator[user.id] = user;

          redisMulti.set(this.REDIS_KEY.USER_INFO(user.id, 'PROFILE'), JSON.stringify(user), 'EX', this.SECONDS_FROM_HOUR);

          return accumulator;
        }, {} as Record<string, IUserProfileRedis>)
      );

    await redisMulti.exec();

    return users;
  }

  //====================FRIEND_LIST====================
  async friendList(userId: string) {
    const friendList = await this.redis.smembers(this.REDIS_KEY.USER_INFO(userId, 'FRIEND'));
    if (friendList.length === 0) return await this.syncFriendDbToRedis(userId);

    friendList.splice(0, 1);

    return friendList;
  }

  async syncFriendDbToRedis(syncUserId: string) {
    const userIdFriend = await this.dataSource
      .getRepository(FriendEntity)
      .find({
        where: {
          user_id: syncUserId,
          type: FRIEND_TYPE.FRIEND,
        },
        order: {
          user_friend_id: 'ASC',
        },
      })
      .then((friend) =>
        friend.reduce(
          (result, item) => {
            result.push(item.user_friend_id);

            return result;
          },
          ['0']
        )
      );

    await this.redis.sadd(this.REDIS_KEY.USER_INFO(syncUserId, 'FRIEND'), userIdFriend);
    await this.redis.expire(this.REDIS_KEY.USER_INFO(syncUserId, 'FRIEND'), this.SECONDS_FROM_HOUR);

    userIdFriend.splice(0, 1);

    return userIdFriend;
  }

  //====================CONVERSATION_DISABLE_NOTIFY_LIST====================
  async conversationDisableNotifyList(userId: string) {
    const conversationDisableNotify = await this.redis.smembers(this.REDIS_KEY.USER_INFO(userId, 'CONVERSATION_DISABLE_NOTIFY'));
    if (conversationDisableNotify.length === 0) return await this.syncConversationDisableNotifyDbToRedis(userId);

    conversationDisableNotify.splice(0, 1);

    return conversationDisableNotify;
  }

  async syncConversationDisableNotifyDbToRedis(userId: string) {
    const conversationDisableNotify = await this.dataSource
      .getRepository(ConversationUserDisableNotifyEntity)
      .find({
        select: ['conversation_id'],
        where: {
          user_id: userId,
        },
      })
      .then((data) =>
        data.reduce(
          (result, item) => {
            result.push(item.conversation_id);
            return result;
          },
          ['0']
        )
      );

    this.redis.sadd(this.REDIS_KEY.USER_INFO(userId, 'CONVERSATION_DISABLE_NOTIFY'), conversationDisableNotify);
    this.redis.expire(this.REDIS_KEY.USER_INFO(userId, 'CONVERSATION_DISABLE_NOTIFY'), this.SECONDS_FROM_HOUR);

    conversationDisableNotify.splice(0, 1);

    return conversationDisableNotify;
  }

  //====================CONVERSATION_PIN_LIST====================
  async conversationPinList(userId: string) {
    const pinnedConversations = await this.redis
      .get(this.REDIS_KEY.USER_INFO(userId, 'CONVERSATION_PINNED'))
      .then((result) => JSON.parse(result) as string[]);
    if (!pinnedConversations) return await this.synConversationPinDbToRedis(userId);

    return pinnedConversations;
  }

  async synConversationPinDbToRedis(userId: string) {
    const conversationPin = await this.dataSource
      .getRepository(ConversationUserPinnedEntity)
      .find({
        select: ['id', 'conversation_id'],
        where: {
          user_id: userId,
        },
        order: {
          id: 'DESC',
        },
      })
      .then((result) => result.map((item) => item.conversation_id));

    await this.redis.set(
      this.REDIS_KEY.USER_INFO(userId, 'CONVERSATION_PINNED'),
      JSON.stringify(conversationPin),
      'EX',
      this.SECONDS_FROM_HOUR
    );

    return conversationPin;
  }

  //====================CONVERSATION_BACKGROUND_LIST====================
  async conversationBackgroundList(userId: string, conversationIdList: string[]) {
    const redisKeyConversationBackgroundList = conversationIdList.map((conversationId) =>
      this.REDIS_KEY.CONVERSATION_USER_BACKGROUND(userId, conversationId)
    );
    const conversationBackgroundList = await this.redis.mget(...redisKeyConversationBackgroundList);

    const result: {
      [conversation_id: string]: object;
    } = {};

    await Promise.all(
      conversationBackgroundList.map(async (item, index) => {
        const conversationBackground = JSON.parse(item);
        if (!conversationBackground) {
          result[conversationIdList[index]] = await this.synConversationBackgroundDbToRedis(userId, conversationIdList[index]);
        } else {
          result[conversationIdList[index]] = conversationBackground;
        }
      })
    );

    return result;
  }

  async conversationUserBackground(conversationId: string, userIds: string[]) {
    const redisKeys = userIds.map((userId) => this.REDIS_KEY.CONVERSATION_USER_BACKGROUND(userId, conversationId));

    const conversationUserBackgrounds = await this.redis.mget(...redisKeys);

    const result: Record<string, IMedia> = {};

    for (let i = 0; i < conversationUserBackgrounds.length; i++) {
      const element = JSON.parse(conversationUserBackgrounds[i]);

      if (!element) {
        result[userIds[i]] = await this.synConversationBackgroundDbToRedis(conversationId, userIds[i]);
      } else {
        result[userIds[i]] = element;
      }
    }

    return result;
  }

  async synConversationBackgroundDbToRedis(userId: string, conversationId: string) {
    const conversationBackground = await this.dataSource.getRepository(ConversationUserBackgroundEntity).findOne({
      select: ['background'],
      where: {
        user_id: userId,
        conversation_id: conversationId,
      },
    });
    if (!conversationBackground) return;

    this.redis.set(
      this.REDIS_KEY.CONVERSATION_USER_BACKGROUND(userId, conversationId),
      JSON.stringify(conversationBackground.background),
      'EX',
      this.SECONDS_FROM_HOUR
    );

    return conversationBackground.background;
  }

  //====================USERS_ONLINE_STATUS====================
  async usersOnlineStatus(userIds: string[]) {
    const keysMulti = this.redis.multi();
    for (const userId of userIds) {
      keysMulti.keys(this.REDIS_KEY.USER_CONNECTION(userId, '*'));
    }
    return await keysMulti.exec().then((data) =>
      data.reduce((accumulator, [error, result], currentIndex) => {
        accumulator[userIds[currentIndex]] = (result as string[]).length ? BOOLEAN.TRUE : BOOLEAN.FALSE;

        return accumulator;
      }, {})
    );
  }

  //====================CONVERSATION_HIDDEN_LIST====================
  async conversationHiddenList(userId: string) {
    const conversationHiddenIds = await this.redis
      .get(this.REDIS_KEY.USER_INFO(userId, 'CONVERSATION_HIDDEN'))
      .then((result) => JSON.parse(result) as string[]);
    if (!conversationHiddenIds) return await this.syncConversationHiddenDbToRedis(userId);

    return conversationHiddenIds;
  }

  async syncConversationHiddenDbToRedis(userId: string) {
    const conversationHidden = await this.dataSource
      .getRepository(ConversationUserHiddenEntity)
      .find({
        select: ['conversation_id'],
        where: {
          user_id: userId,
        },
      })
      .then((result) => result.map((item) => item.conversation_id));

    await this.redis.set(
      this.REDIS_KEY.USER_INFO(userId, 'CONVERSATION_HIDDEN'),
      JSON.stringify(conversationHidden),
      'EX',
      this.SECONDS_FROM_HOUR
    );

    return conversationHidden;
  }
}
